# Contributing to Lektor

- [License](#license)
- [Building](#buidling)
- [Debugging lektor](#debuging)
- [Style and code](#style-and-code)
- [Merge request titles](#merge-request-titles)
- [Modules](#modules)
- [Command table](#command-table)

---

Clone the [git repository](https://git.iiens.net/martin2018/lektor). Create a
branch from master with the `git checkout -b dev-${pseudal}` command. Please,
rebase regularly your branch on `master`.

Please, add your name in the CONTRIBUTORS file. The format is `First name
'Pseudal' FAMILY　NAME`. The use license is the very permisive ISC license.

## License

See [LICENSE](LICENSE).

## Building

It is recommended to use the `../configure` method to build while in
developpement. Also, prefere useing the `--with-depends` with the
configuration, to use the fixed version.

An archlinux package exists, use `makepkg -si` to build and install it. If
there are no package for your distribution, feel free to add a way to make it.

Here are some possible invocations of the `configure` script:

- `../configure --prefix=$HOME/.local --with-liblektor=static --enable-debug CC=clang`
- `../configure --prefix=$HOME/.local --enable-klkt CC=clang CXX=clang++`
- `../configure --prefix=$HOME/.local CC=gcc`
- `../configure --prefix=$HOME/.local --with-depends`

## Debuging

You will need gdb. You can launch gdb on installed programs, gut you can also
use libtool for that.  With the `libtool-bin` package installed, you can run:
`libtool --mode=execute gdb $BUILD_DIR/src/lektord` to launch lektord with gdb.

## Style and the code

### C code and style

Prefer using `__${name}` names for static variables and functions. Prefix the
name of the file or folder in function names, e.g. `database_` or `command_`.
If you declare functions that really nobody should used, please prefix them
with three underscores like `___${name}`.

For formating, use the `./utils/scripts/astyle.bash` from the root of the
project. You'll need [astyle](http://astyle.sourceforge.net/) for that. Note
that the astyle step is a part of the CI/CD.

You are encouraged to use the `./utils/scripts/tags.bash` script to build the
`ctags` and `cscope` database files to explore more easelly the project.

The C standard used here is C11, because on debian C18 doesn't compiles. The
C++ used standard is C++17, thus C++ code is currently not tested with the CI.

### Scripts

Scripts goes to the [scripts](utils/scripts) folder. They must be written to be
launched from the root of the project. Include the language used in the name of
the script, for example:

- `astyle.bash`: written in bash (bourn shell again)
- `hashire.sh`: written in sh (bourn shell)
- `magic.rb`: written in ruby

Please, do not write scripts in too many different languages to reduce the
number of dependencies. If you are using binaries in your scripts (like astyle
or bc), include them in the binary dependencies in the [README](README.md)
file. Also note that `bash` and `sh` are two different shells and things that
work in `bash` might not work with `sh`. As `bash` usually provide `sh`, it is
fine to use the two shells in different scripts without increasing the
dependency number.

There is a strong possibility that your script will do similar things to what a
script already in this repo do. In that case, please work on the existing
script and enhance it with your changes. No need to duplicate things.

### Utilities

You can place utilities in the [utils](utils) folder. When downloading or
building tools not dirrectly distributed with lektor this folder will be used.

## Merge request titles

Try to be as more descriptive as you can in your Merge Request title. Include
the issue number and the target OS if possible.

- Branch from the master branch and, if needed, rebase to the current master
  branch before submitting your merge request. If it doesn't merge cleanly with
  master you may be asked to rebase your changes
- Commits should be as small as possible, while ensuring that each commit is
  correct independently (i.e., each commit should compile and pass tests)
- Add tests relevant to the fixed bug or new feature. In general, add a way to
  test lektor

When adding new fetures or solving issues / bugs, please edit the
[CHANGELOG](CHANGELOG.md) file at the `# Next` section. This will help people
know what has been done since the last version.

## Modules

Modules are a way of putting modular code in lektor. It enable the user to
select different types of players and repositories.

Modules are `.so` files (i.e. dynamic libraries) loaded by lektors using a
function named `module_set_function` which takes a `void *` and a
`struct lkt_option *` in argument and must return an `int` (may return an
`uint16_t` in the future). This function must initialize correctly the module.
The name of this function can be defined in the INI file.

The functions of a modules are stored in a `register`. This register can be
registered in the lektor's global `register` or exported inside `.so` file.
Only one `register` can be exported by `.so` file. The macros are here to help
you do so.

Some functions are mendatory for a module:

- The `new(void **)` function creates the module
- The `free(void **)` release the module, it *can't* be used after
- The `close(void **)` close the module, it *can* be used after

All exported functions must take a `va_list`. The first argument of the `va_list`
is a `void **`, which point to the pointer used to access the module's defined
structure. The rest is up to you, no checks will be done. All functions must return
an `int` in a POSIX way: `0` if everything is ok, an other value if something
went wrong.

## Command table

Here is a list of commands used / not used / not implemented in lektord. If the
command differs from what MPD do, it is also specified.

### MPD command compatibility table

For the compatibility column, the possible values are the following:

- `-`: has a poor compatibility, doesn't work as MPD intended it
- `~`: basic compatibility, some arguments types / number may change
- `+`: is compatible
- `CHECK`: check for compatibility status
- `FIXME`: was intended to work as in MPD but is not, fix it
- `TODO`: marked as 'to be implemented', check the next releases if it is implemented
- nothing: has not been implemented and is not a priority

| MPD command                               | lektord command               | compatibility | compatibility note                                |
| ----------------------------------------- | ----------------------------- | :-----------: | ------------------------------------------------- |
| `clearerror`                              |                                       |       | not implemented                                   |
| `currentsong`                             | `currentsong`                         | ~     | use the lektord tags, not MPD once                |
| `idle [subsistem]`                        | `idle [subsistem]`                    | +     |                                                   |
| `status`                                  | `status`                              | ~     | slight differencies in the duration and time, not working with mpc |
| `stats`                                   | `stats`                               | ~     | db\_playtime and playtime files are missing       |
| `consume {state}`                         | `consume {state}`                     | +     |                                                   |
| `random {state}`                          | `random {state}`                      | +     |                                                   |
| `repeat {state}`                          | `repeat {state}`                      | +     |                                                   |
| `setvol {vol}`                            | `setvol {vol}`                        | +     | won't display the correct volume, mpv properties flush issue... |
| `getvol`                                  | `getvol`                              | +     | will display the correct value after the end of the kara |
| `single {state}`                          | `single {state}`                      | +     |                                                   |
| `replay_gain_mode {mode}`                 |                                       |       | not implemented                                   |
| `replay_gain_status`                      |                                       |       | not implemented                                   |
| `next`                                    | `next`                                | +     |                                                   |
| `pause {state}`                           | `pause {state}`                       | +     |                                                   |
| `play [songpos]`                          | `play [songpos]`                      | +     | if no `songpos` passed, defaults to `0`           |
| `playid [songid]`                         | `playid {songid}`                     | +     | get the first found `songid` in the queue         |
| `previous`                                | `previous`                            | +     |                                                   |
| `seek {songpos} {time}`                   | `seek {songpos} {time}`               | +     |                                                   |
| `seekid {songid} {time}`                  | `seekid {songid} {time}`              | +     | seek to the first found id, sqlite dependent      |
| `seekcur {time}`                          | `seekcur {time}`                      | +     |                                                   |
| `stop`                                    | `stop`                                | +     |                                                   |
| `add {uri}`                               | `add {uri}`                           | +     |                                                   |
| `addid {id} [pos]`                        | `addid {id...}`                       | ~     | can add multiple songs, but no position support   |
| `clear`                                   | `clear`                               | +     |                                                   |
| `delete [{pos}/{start:end}]`              | `delete {pos}`                        | -     | can take only one position                        |
| `deleteid {songid}`                       | `deleteid {songid}`                   | +     | delete all kara with that id, skip the current kara if needed |
| `move [{from}/{start:end}] {to}`          | `move {from} {to}`                    | ~     | `{to}` is always an absolute position             |
| `moveid {from} {to}`                      |                                       |       |                                                   |
| `playlist`                                | `playlist`                            | +     |                                                   |
| `playlistfind {tag} {needle}`             | `playlistfind {uri}`                  | ~     | uses lektord URIs                                 |
| `playlistid {songid}`                     | `playlistid {id}`                     | ~     | the uri is only an id                             |
| `playlistinfo [[songpos]/[start:end]]`    | `playlistinfo`                        | -     | is an alias to `playlist`                         |
| `playlistsearch {taf} {needle}`           | `playlistsearch {uri}`                | ~     | uses lektord URIs                                 |
| `plchanges {version} [start:end]`         |                                       |       | not implemented                                   |
| `plchangesposid {version} [start:end]`    |                                       |       | not implemented                                   |
| `prio {prio} {start:end}`                 |                                       |       | not implemented                                   |
| `prioid {prio} {id...}`                   |                                       |       | not implemented                                   |
| `rangeid {id} {start:end}`                |                                       |       | not implemented                                   |
| `shuffle [start:end]`                     | `shuffle`                             | ~     | shuffles all the queue                            |
| `swap {song1} {song2}`                    | `swap {song1} {song2}`                | +     |                                                   |
| `swap {songid1} {songid2}`                |                                       |       | not implemented                                   |
| `listplaylist {name}`                     | `listplaylist {name}`                 | +     |                                                   |
| `listplaylistinfo {name}`                 | `listplaylistinfo {name} {uri}`       | -     | do a search in the playlist                       |
| `listplaylists`                           | `listplaylists`                       | +     |                                                   |
| `load {name} [start:end]`                 |                                       |       | not implemented, use `add {uri}` to do that       |
| `playlistadd {name} {uri}`                | `playlistadd {name} {uri}`            | ~     | uses lektord URIs                                 |
| `playlistclear {name}`                    | `playlistclear {name}`                | +     |                                                   |
| `playlistdelete {name} {songpos}`         | `playlistdelete {name} [songid]`      | ~     | delete all the playlist if `songpos` is not spesified, uses kara ids, not pos |
| `rename {name} {newname}`                 | `rename {name} {newname}`             | ~     | will fail if the `newplaylist` already exists     |
| `rm {name}`                               | `rm {name}`                           | +     |                                                   |
| `save {name}`                             | `save {name} {destdb}`                | -     | export a part of the db into as sqlite3 db        |
| `count {filt} [group {grptype}]`          |                                       |       | not implemted                                     |
| `getfingerprint {uri}`                    |                                       |       | not implemted                                     |
| `find {filt} [sort {type}] [window {START:END}]`      | `find {uri}`              | -     | do a simple search in the db with lektord's URIs  |
| `findadd {filt} [sort {type}] [window {start:end}]`   | `findadd {uri}`           | -     | same remarks as `find`                            |
| `list {type} {filt} [group {grptype}]`    | `list {uri}`                          | -     | alias to `find`                                   |
| `listall [uri]`                           | `listall {uri}`                       | -     | alias to `find`                                   |
| `listallinfo [uri]`                       | `listallinfo {uri}`                   | -     | alias to `find`                                   |
| `listfiles {uri}`                         | `listfiles {uri}`                     | -     | alias to `find`                                   |
| `lsinfo [uri]`                            | `lsinfo {uri}`                        | -     | alias to `find`                                   |
| `search {filt} [sort {type}] [window {start:end}]`    | `search {uri}`            | -     | is an alias to `find`                             |
| `searchadd {filt} [sort {type}] [window {start:end}]` | `searchadd {uri}`         | -     | is an alias to `findadd`                          |
| `update [uri]`                            | `update`                              | ~     | doesn't take URIs                                 |
| `rescan [uri]`                            | `rescan`                              | ~     | doesn't take URIs                                 |
| `close`                                   | `close`                               | +     | send the `OK` status before closing the socket    |
| `kill`                                    | `kill`                                | +     |                                                   |
| `password {password}`                     | `password {password}`                 | +     |                                                   |
| `ping`                                    | `ping`                                | +     |                                                   |
| `config`                                  | `config`                              | +     | all authentified client can do this command       |
| `sticker get {t} {uri} {n}`               | `sticker get {type} {uri} {name}`     | CHECK | check the return conventions                      |
| `sticker set {t} {uri} {n} {val}`         | `sticker set {t} {uri} {n} {val}`     | CHECK | check the return conventions                      |
| `sticker delete {t} {uri} [n]`            | `sticker delete {t} {uri} [n]`        | CHECK | check the return conventions                      |
| `sticker list {type} {uri}`               | `sticker list {type} {uri}`           | CHECK | check the return conventions                      |
| `sticker find {t} {uri} {n}`              | `sticker find {t} {uri} {n}`          | CHECK | check the return conventions                      |
| `sticker find {t} {uri} {n} {op} {v}`     | `sticker find {t} {uri} {n} {op} {v}` | CHECK | check the return conventions, operators are `g`, `e` and `l` |

### Commands specific to lektord

| Command                 | Comment / notes                                                 |
| ----------------------- | --------------------------------------------------------------- |
| `__import`              | import favorites as playlists prefixed bt `@` in lektor         |
| `__dump {pltname}`      | add the content of the queue to an existing playlist            |
| `__adduser {password}`  | create a new admin user                                         |
| `__restart`             | restarts lektord                                                |
| `__rescan`              | same as `rescan`, but force populating without timestamp check  |
| `sticker __add {name}`  | create a sticker                                                |
| `__insert {uri}`        | do the `add` command, but with the maximum priority             |
| `__flat`                | flat is justice => all prio are now 1 in the queue              |

### Commands with no plan for support

| MPD command                                                       | Reason / notes                                |
| ----------------------------------------------------------------- | --------------------------------------------- |
| `readcomments {uri}`                                              | really not a priority                         |
| `readpicture {uri} {offset}`                                      | really not a priority                         |
| `albumart {uri} {offset}`                                         | really not a priority                         |
| `mount {path} {uri}`                                              | no necessity in lektord                       |
| `unmount {path}`                                                  | no necessity in lektord                       |
| `listmounts`                                                      | no necessity in lektord                       |
| `listneighbors`                                                   | no necessity in lektord                       |
| `partition {name}`                                                | no necessity in lektord                       |
| `listpartitions`                                                  | no necessity in lektord                       |
| `newpartition {name}`                                             | no necessity in lektord                       |
| `delpartition {name}`                                             | no necessity in lektord                       |
| `tagtypes`                                                        | tags are handled by kurisu for the moment     |
| `tagtypes disable {name...}`                                      | tags are handled by kurisu for the moment     |
| `tagtypes enable {name...}`                                       | tags are handled by kurisu for the moment     |
| `tagtypes clear`                                                  | tags are handled by kurisu for the moment     |
| `tagtypes all`                                                    | tags are handled by kurisu for the moment     |
| `addtagid {songid} {tag} {val}`                                   | tags are handled by kurisu for the moment     |
| `cleartagid {songid} [tag]`                                       | tags are handled by kurisu for the moment     |
| `crossfade {seconds}`                                             | lektord uses libmpv                           |
| `mixrampdb {dB}`                                                  | lektord uses libmpv                           |
| `mixrampdelay {seconds}`                                          | lektord uses libmpv                           |
| `moveoutput {outputname}`                                         | lektord uses libmpv                           |
| `disableoutput {id}`                                              | lektord uses libmpv                           |
| `enableoutput {id}`                                               | lektord uses libmpv                           |
| `toggleoutput {id}`                                               | lektord uses libmpv                           |
| `outputs`                                                         | lektord uses libmpv                           |
| `outputset {id} {name} {value}`                                   | lektord uses libmpv                           |
| `commands`                                                        | not implemented                               |
| `notcommands`                                                     | not implemented                               |
| `urlhandlers`                                                     | not implemented                               |
| `decoders`                                                        | lektord only suprts mkv files (for mkv tags)  |
| `subscribe {name}`                                                | no necessity for p2p messages for the moment  |
| `unsubscribe {name}`                                              | no necessity for p2p messages for the moment  |
| `channels`                                                        | no necessity for p2p messages for the moment  |
| `readmessages`                                                    | no necessity for p2p messages for the moment  |
| `sendmessage {channel} {text}`                                    | no necessity for p2p messages for the moment  |
| `searchaddpl {name} {filt} [sort {type}] [window {start:end}]`    | `playlistadd` exists                          |
| `playlistmove {name} {from} {to}`                                 | no necessity in lektord as playlists are shuffled when added to the queue |
