#define _POSIX_C_SOURCE 200809L

#include <lektor/lktconfig.h>
#include <lektor/cmd.h>
#include <lektor/common.h>
#include <sys/types.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

const char *executable_name = NULL;

noreturn void
print_help(void)
{
    if (!executable_name)
        exit(EXIT_FAILURE);
    const char *const args[] = { LKT_MAN_BINARY, executable_name, NULL };
    execv(args[0], (char *const *) args);
    exit(EXIT_FAILURE);
}

noreturn void
cmd_parse(struct cmd_opt *opts, int argc, const char **argv)
{
    int count = 0, is_ok, offset = 0;
    struct cmd_opt *it = opts;
    cmd_callback call[2] = { NULL, NULL };

    if (argc == 0 || *argv == NULL || ! strcasecmp(argv[0], "help") ||
        ! strcasecmp(argv[0], "--help") || ! strcasecmp(argv[0], "-help"))
        goto help;

    /* Find the command */
    while (it && it->name) {
        is_ok = (! strncasecmp(argv[0], it->name, strlen(argv[0])));
        call[is_ok] = it->call;
        count += is_ok;
        it = opts + (++offset);
    }

    /* Now search for a unique match */
    if (count > 1)
        goto not_exclusive;

    if (!call[1] || count == 0)
        goto not_found;

    struct cmd_args arguments = {
        .argc = argc - 1,
        .argv = (const char **) & (argv[1]),
    };

    /* This call should exit. */
    call[1](&arguments);
    exit(EXIT_FAILURE);

not_found:
    /* The default function */
    if (!it->name && it->call) {
        struct cmd_args arguments = {
            .argc = argc,
            .argv = (const char **) argv,
        };
        it->call(&arguments);
    }

    LOG_ERROR("COMMAND", "Command '%s' could not be found", argv[0]);
    exit(EXIT_FAILURE);

help:
    print_help();

not_exclusive:
    LOG_ERROR("COMMAND", "Failed to determine option, '%s' not exclusive",
              argv[0]);
    exit(EXIT_FAILURE);
}
